# Copyright (c) 2013-2019 Gerald Knizia
# 
# This file is part of the 'cc_eqs_pv' program
# (See https://sites.psu.edu/knizia/)
# 
# cc_eqs_pv is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# cc_eqs_pv is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with cc_eqs_pv (LICENSE_GPL_v3.txt). If not, see
# http://www.gnu.org/licenses/

"""low level functions for emitting sets of C++ code."""
from __future__ import print_function
from textwrap import dedent

_OutputLines = []
_Indent = ""

def Emit(s="",*FormatArgs):
   """write to _OutputLines. Saved to file when Dump(...) is called."""
   if FormatArgs:
      s = s.format(*FormatArgs)
   global _OutputLines
   for Line_ in s.split("\n"):
      Line = Line_.rstrip()
      if Line != "":  # <- don't indent empty lines.
         Line = _Indent + Line
      _OutputLines.append(Line)

def Comment(s,*FormatArgs):
   if FormatArgs:
      s = s.format(*FormatArgs)
   xs = s.split("\n",1)
   if len(xs) == 1:
      Emit("// %s" % s)
   else:
      s = xs[0] + "\n" + dedent(xs[1])
      for Line in s.splitlines():
         Emit("// %s" % Line)

def Ind():
   """raise indent on following lines"""
   global _Indent
   _Indent = _Indent + "   "

def UnInd():
   """lower indent on following lines"""
   global _Indent
   _Indent = _Indent[:-3]

def Dump(FileName):
   """write current code to FileName and reset"""
   global _OutputLines
   File = open(FileName, "w")
   Text = "\n".join(_OutputLines) + "\n"
   Text = GetHeader(FileName) + "\n" + Text
   File.write(Text)
   File.close()
   _OutputLines = []

_NumberChars = "0123456789" + "abcdefghijklmnopqrstuvwxyz" + "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
def NumberChars(x):
   import numpy as np
   """used to form hacky names of recursion intermediates with short names.  E.g.,
   something with angular momenta 132 might be cast into r_132. But this will lead
   to problems if any of the numbers is larger than 9.... so we put in some more
   other letter-like stuff. If this becomes a real problem, one could use one of
   the characters as an escape sequence."""
   if isinstance(x,np.ndarray) or isinstance(x,tuple):
      return "".join(map(NumberChars, x))
   return _NumberChars[x]

_LastPrototype = None
def BeginFn(Prototype,*FormatArgs):
   Emit(Prototype, *FormatArgs)
   Emit("{")
   Ind()
   global _LastPrototype
   _LastPrototype = Prototype

def EndFn():
   UnInd()
   Emit("}\n")
   return _LastPrototype + ";"

def BeginBlock(s):
   Emit(s + " {");
   Ind()

def EndBlock(s=None):
   UnInd()
   Emit("}" + (s if s else ""))

def BeginDecl(s):
   Emit(s)
   Ind()

def EndDecl(s=None):
   UnInd()

def EmitTable(Type, Name, Data, FmtFn=None, EntriesPerLine=8, EntrySize=None, Desc=None):
   if Desc:
      Emit()
      Comment(Desc)
   BeginBlock("{} {}[{}] =".format(Type,Name,len(Data)))
   RestData = tuple(Data)
   if FmtFn is None:
      FmtFn = lambda s: str(s)
   while RestData:
      nEntries = min(len(RestData), EntriesPerLine)
      LineData = RestData[:nEntries]
      RestData = RestData[nEntries:]
      Line = ", ".join(FmtFn(o) for o in LineData)
      if RestData:
         Line = Line + ","
      Emit(Line)
   if EntrySize:
      EndBlock("; // %.2f kb" % ((1./1024)*(len(Data) * EntrySize)))
   else:
      EndBlock(";")

def FloatToStr1(f):
   if 1:
      # yes, I know...
      for c in range(-15,16):
         for d in range(1,16):
            cd = c/(1.*d)
            if abs(f - cd)<1e-10: f = cd
            if abs(f + cd)<1e-10: f = -cd
      # c/p'd from VtfAsm
      s = "%.15e" % f
      # ^- FIXME: I don't get it anymore...
      # (hm.. i think this is supposed to turn rational numbers into exact rationals)
      # unfortunately, it is very slow...
   else:
      if abs(f-1.)<1e-10: f = 1.
      if abs(f+1.)<1e-10: f = -1.
   s = "%.15e" % f

   ie = s.index('e')
   ilast = ie-1
   while s[ilast] == '0':
      ilast -= 1
   s = s[:ilast+1] + s[ie:]
   if ( s.endswith("e+00") ):
      s = s[:-4]
   #print >> stderr, "Float2Str: f=",f,"  ->  ",s
   #print("Float2Str: f=",f,"  ->  ",s, file=sys.stderr)
   return s

def FloatToStr(f):
   return "%r" % f
   return FloatToStr1(f)

def GetHeader(FileName):
   from pwd import getpwnam
   from time import tzname, localtime
   from socket import gethostname
   from getpass import getuser
   lt = localtime()
   return "/* %s v%4i%02i%02i %s [%s, %s] */"  % (FileName, lt[0], lt[1], lt[2], tzname[0], gethostname(), getpwnam(getuser()).pw_gecos)

if __name__ == "__main__":
   print(GetHeader("IrAmrr.cpp"))

