# coding: utf-8
# Copyright (c) 2013-2019 Gerald Knizia
# 
# This file is part of the 'cc_eqs_pv' program
# (See https://sites.psu.edu/knizia/)
# 
# cc_eqs_pv is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# cc_eqs_pv is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with cc_eqs_pv (LICENSE_GPL_v3.txt). If not, see
# http://www.gnu.org/licenses/

""" This program generates closed-shell Coupled Cluster equations as described in

   Cong Wang, Gerald Knizia - A simple permutation group approach to spin-free
   higher-order coupled cluster methods (2018)
   https://arxiv.org/abs/1805.00565

It is released as accompanying information to this article. The source code itself
is released under the GNU General Public License v2 (see
https://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html).

Author: Gerald Knizia, 2013-2019

Notes on usage:
- It requires a reasonably current version of SymPy to be installed, as it uses
  sympy's Permutation and PermutationGroup objects. This version is tested with
  sympy 1.1.2dev

- It runs faster on pypy, due to using lots of low level index manipulations, which
  are intrinsically very slow in interpreted Python. On linux on fedora, it can
  be run, for example, as
  
  PYTHONPATH=/usr/lib/python2.7/site-packages/ pypy cc_eqs_pv.py
  
  or if only python3-sympy is available, as
  
  PYTHONPATH=/usr/lib/python3.7/site-packages/ pypy cc_eqs_pv.py  
  
  or on Windows 10 with WSL's version of Ubuntu:
  
  PYTHONPATH=/usr/lib/python3/dist-packages/ pypy cc_eqs_pv.py
  
  In either case, where the first argument sets the path to the installed sympy
  installation (you should specify the directory to whereever you installed or
  downloaded sympy to. Note that sympy is a pure python library and does not
  need to be compiled or configured, just downloaded and unpacked.)

- The program by default just generates sets of equation lists in the "eqs"
  subdirectory (formated as .itfaa --- ITF Algorithm Assembler contraction files).
  If invoked with the command line argument
  
   --export <dir>
   
  it will also generate a set of SRCI inline files which encode the equation lists
  as C++ data elements for numerical testing in SRCI.

- The actual methods generated are specified at the very end of this file.
  If no --export is used, the program has no external dependencies apart from
  sympy.
  
"""

from __future__ import print_function
import itertools as it
import sys # for sys.argv (command line arguments to this program)
from time import time
from collections import namedtuple

from sympy.combinatorics import Permutation, PermutationGroup
from sympy.combinatorics.permutations import _af_rmul, _af_rmuln, _af_invert, _af_new
from sympy.combinatorics.util import (_distribute_gens_by_base, _orbits_transversals_from_bsgs)


# occupation number of an occupied orbital.
ORB_OCC = 2


import platform
_ThisIsPython2 = (int(platform.python_version_tuple()[0]) == 2)


if not _ThisIsPython2:
   def cmp(a,b):
      return (a > b) - (a < b)

def _Factorial(n):
   if n <= 1:
      return 1
   else:
      return n * _Factorial(n-1)

FIndexedTensor = namedtuple("FIndexedTensor", ["Name", "Indices"])

def _assign_perm_group_bsgs(self, base, strong_gens):
   # c/p from PermGroup.schreier_sims; idea is that we can make a PermutationGroup
   # object with a BSGS supplied from the outside. Reason is that we may need
   # a base which is in lexicographical order, and schreier_sims does not necessarily
   # make one of this kind as implemented in sympy currently.
   base, strong_gens = self.schreier_sims_incremental()
   self._base = base
   self._strong_gens = strong_gens
   if not base:
      self._transversals = []
      self._basic_orbits = []
      return

   strong_gens_distr = _distribute_gens_by_base(base, strong_gens)
   basic_orbits, transversals = _orbits_transversals_from_bsgs(base,\
            strong_gens_distr)
   self._transversals = transversals
   self._basic_orbits = [sorted(x) for x in basic_orbits]


def _is_minimal_bsgs(base, gens):
   """check if bsgs is minimal under lexicographical order."""
   base1 = []
   sgs1 = gens[:]
   size = gens[0].size
   for i in range(size):
      if not all(h._array_form[i] == i for h in sgs1):
         base1.append(i)
         sgs1 = [h for h in sgs1 if h._array_form[i] == i]
   return base1 == base


def _get_minimal_bsgs(base, gens):
   """given a BSGS, transform it to a BSGS minimal under lexicographical order
   by using base swaps."""
   G = PermutationGroup(gens)
   base, gens = G.schreier_sims_incremental()
   # cgk: I inserted stuff to re-sort the base.
   base_sorted = list(sorted(base))
   N = len(base)
   while 1:
      Swapped = False
      for i in range(N-1):
         if base[i+1] < base[i]:
            base, gens = G.baseswap(base, gens, i)
            Swapped = True
      if not Swapped:
         break
   if not _is_minimal_bsgs(base, gens):
      return None
   return base, gens


def _make_group_with_minimal_bsgs(n, Gens):
   """generates a PermutationGroup object from a given set of generators,
   and, if necessary, generates a minimal BSGS for it."""
   def FmtIndices(L):
      return u"".join("%s"%o for o in L)
   def FmtT(Ti):
      return u"%s[%s]" % (Ti[0], FmtIndices(Ti[1]))
   if not Gens:
      PermGroup = PermutationGroup([Permutation(size=n)])
   else:
      Gens = [Permutation(o, size=n) for o in Gens]
      mbsgs = _get_minimal_bsgs(range(Gens[0].size), Gens)
      assert(mbsgs is not None)
      PermGroup = PermutationGroup(mbsgs[1])
      _assign_perm_group_bsgs(PermGroup, mbsgs[0], mbsgs[1])
      if 0:
         # some tests; went all through with CCSDT and full symmetries.
         assert(PermGroup.base == mbsgs[0])
         assert(PermGroup.strong_gens == mbsgs[1])
         PermGroup2 = PermutationGroup(Gens)
         assert(PermGroup2.is_subgroup(PermGroup))
         assert(PermGroup.is_subgroup(PermGroup2))
   return PermGroup


def IsPermutation(L):
   """returns whether the sequence L contains the numbers 0..L-1 (and nothing else)"""
   N = len(L)
   for i in L:
      assert(isinstance(i, int))
      if i < 0 or i >= N:
         return False
   if len(set(L)) != len(L):
      return False
   return True


def FindCycles(l,IncludeLength1=True):
   """return the cycles a permutation is composed of"""
   # didn't want to bother making my own... c/p'd from
   # http://corte.si/posts/code/cyclesort/index.html
   seen = set()
   cycles = []
   for i in range(len(l)):
      if (not i in seen) and (IncludeLength1 or i != l[i]):
         cycle = []
         n = i
         while 1:
            cycle.append(n)
            n = l[n]
            if n == i:
               break
         seen = seen.union(set(cycle))
         cycles.append(list(reversed(cycle)))
   return cycles


def FloatToStr1(f):
   if 1:
      # yes, I know...
      for c in range(-15,16):
         for d in range(1,16):
            cd = c/(1.*d)
            if abs(f - cd)<1e-10: f = cd
            if abs(f + cd)<1e-10: f = -cd
      # c/p'd from VtfAsm
      s = "%.15e" % f
      # ^- FIXME: I don't get it anymore...
      # (hm.. i think this is supposed to turn rational numbers into exact rationals)
      # unfortunately, it is very slow...
   else:
      if abs(f-1.)<1e-10: f = 1.
      if abs(f+1.)<1e-10: f = -1.
   s = "%.15e" % f

   ie = s.index('e')
   ilast = ie-1
   while s[ilast] == '0':
      ilast -= 1
   s = s[:ilast+1] + s[ie:]
   if s.endswith("e+00"):
      s = s[:-4]
   #print >> stderr, "Float2Str: f=",f,"  ->  ",s
   return s


def FloatToStr(f):
   return "%r" % f
   return FloatToStr1(f)



# See comments in GetLeftPermsAndFactors. These are made by covariant_proj.py, and stored
# here so that numpy is not required for running the program.
#
# Note on the denominators: They are factorial(2*n-1), where n is the substitution degree.
_DefaultContravariantLeftPermAndFactors = {
   1: [((0,), 0.5)],
   2: [((0, 1), 1./3.), ((1, 0), 1./6.)],
   #2: [((0, 1), 1.)],
   3: [((0, 1, 2), 17./120.), ((0, 2, 1), -1./120.), ((1, 0, 2), -1./120.), ((1, 2, 0), -7./120.), ((2, 0, 1), -7./120.), ((2, 1, 0), -1./120.)],
   4: [((0, 1, 2, 3), 258./5040.), ((0, 1, 3, 2), -48./5040.), ((0, 2, 1, 3), -48./5040.), ((0, 2, 3, 1), -57./5040.), ((0, 3, 1, 2), -57./5040.), ((0, 3, 2, 1), -48./5040.), ((1, 0, 2, 3), -48./5040.), ((1, 0, 3, 2), 78./5040.), ((1, 2, 0, 3), -57./5040.), ((1, 2, 3, 0), 42./5040.), ((1, 3, 0, 2), 42./5040.), ((1, 3, 2, 0), -57./5040.), ((2, 0, 1, 3), -57./5040.), ((2, 0, 3, 1), 42./5040.), ((2, 1, 0, 3), -48./5040.), ((2, 1, 3, 0), -57./5040.), ((2, 3, 0, 1), 78./5040.), ((2, 3, 1, 0), 42./5040.), ((3, 0, 1, 2), 42./5040.), ((3, 0, 2, 1), -57./5040.), ((3, 1, 0, 2), -57./5040.), ((3, 1, 2, 0), -48./5040.), ((3, 2, 0, 1), 42./5040.), ((3, 2, 1, 0), 78./5040.)]
}

def GetLeftPermsAndFactors(OrderR, CoOrContraVariant="contra"):
   """Determine "good" linear combinations of ABC.. permutations on the lhs of
   an expectation value

         <e^{IJK...}_{ABC...} ...>

   such that the number of resulting contracted terms are minimized. For
   example, in the standard closed-shell case, projecting onto

      <(1/6) (2 e^IJ_AB + e^IJ_BA) ...>

   leads to many terms less than direct projection onto <e^IJ_AB ...>,
   since this linear combination is "almost" bi-orthogomal to  <... e^ab_ij>
   on the rhs."""
   assert(CoOrContraVariant in ["co","contra"])
   if CoOrContraVariant == "contra" and OrderR != 0:
      assert(ORB_OCC == 2 and OrderR in _DefaultContravariantLeftPermAndFactors)
      return _DefaultContravariantLeftPermAndFactors[OrderR]
   else:
      # nothing given--make a default arrangement.
      return [(list(range(OrderR)), 1./_Factorial(OrderR))]
   pass


def MakeExpTOpDecl(MaxSubst, TRanks, Commutative=True, LinearOnly=False, MaxOrderT=None):
   """make the combinations of T operators occurring in an
   exponential; if Commutative, also produce prefactors.

   Output is a dictionary mapping {(n0,n1,n2,...) : Factor} which
   represents the term
      += Factor * t0^n0 * t1^n1 * t2^n2 ...
   with t0 = 1, t1 = t^i_a E^a_i, t2 = (1/2) t^ij_ab E^ab_ij
   """
   Factors = {(0,): 1.}
   def ExtendFactors(Base):
      for TRank in TRanks:
         SubstTerm = Base + (TRank,)
         if sum(SubstTerm) > MaxSubst:
            continue
         if Commutative:
            Key1 = tuple(sorted(SubstTerm))
         else:
            Key1 = Key
         Factors[Key1] = Factors.get(Key1, 0.) + 1
         ExtendFactors(SubstTerm)
   ExtendFactors(())
   FactorsOut = {}
   for (Key,Factor) in Factors.items():
      ExpOrder = len(Key)
      Factor /= _Factorial(ExpOrder)
      if LinearOnly and len(Key) > 1:
         continue
      if MaxOrderT is not None and len(Key) > MaxOrderT:
         # skip terms originating from T^k contributions with k > MaxOrder.
         # Note that this refers to orders in T = T1 + T2 + T3 + ..., not the substitution degree.
         continue
      FactorsOut[Key] = Factor
   return FactorsOut


class FIndexDomain(object):
   """represents a subset of the molecular orbital space. Used to represent
   the closed-shell molecular orbital subspace and the external (=unoccupied)
   molecular orbital subspace"""
   def __init__(self, Name, ShortName, BaseIndices, CreatorIndexContractionDirections, SortPriority):
      self.BaseIndices = BaseIndices
      for o in self.BaseIndices:
         assert(o.islower())
         # ^- later on we assume that free and dummy indices can be made
         #    by just upper/lower-casing these guys.
      self.Name = Name
      self.ShortName = ShortName
      self.CreatorIndexContractionDirections = CreatorIndexContractionDirections
      for cicd in self.CreatorIndexContractionDirections:
         # decides whether a creation index of this class in iEop
         # can be contracted to a destruction index of this class
         # in another jEop, provided:
         #     jEop < iEop (Left),
         #     jEop == iEop (Center)
         #     jEop > iEop (Right)
         assert(cicd in ["L", "C", "R"])
      # decides on how far left indices of this domain like to be.
      self.SortPriority = SortPriority

   def __hash__(self):
      return hash(self.Name)

   def __repr__(self):
      return "id-{}".format(self.ShortName)
   
   def MakeIndexName(self, ii):
      if ii < len(self.BaseIndices):
         return self.BaseIndices[ii]
      else:
         ii = ii - len(self.BaseIndices)
         return "{}{}".format(self.BaseIndices[0], ii)
      pass


class FIndexPool(object):
   """maintain a list of indices of a set of index domains which can still be used."""
   def __init__(self, IndexDomains):
      # maps FIndexDomain to number of indices already used.
      self.Free = {}  
      self.Dummy = {}
      for Domain in IndexDomains:
         assert(isinstance(Domain, FIndexDomain))
         self.Free[Domain] = 0
         self.Dummy[Domain] = 0
   
   def TakeFree(self, Domain):
      iNext = self.Free[Domain]
      self.Free[Domain] = iNext + 1
      return Domain.MakeIndexName(iNext).upper()

   def TakeDummy(self, Domain):
      iNext = self.Dummy[Domain]
      self.Dummy[Domain] = iNext + 1
      return Domain.MakeIndexName(iNext).lower()

   def Take(self, Domain, AsFree):
      if AsFree:
         return self.TakeFree(Domain)
      else:
         return self.TakeDummy(Domain)


COP_Create = 1     # create an electron in an orbital
COP_Destroy = -1   # destroy an electron in an orbital

class FOpTerm(object):
   """represents an second-quantized operator (such as T or W) or a placeholder
   for one's derivatives (R). These objects are used to define the target expressions
   to make equations for in operator products."""
   def __init__(self, Name, Factor, DomainSets, SlotSymmetry, SwitchUpperLower=False, LowerPerms=None, ThisIsLhs=False):
      self.SwitchUpperLower = SwitchUpperLower
      self.ThisIsLhs = ThisIsLhs

      self.Name = Name

      self.Factor = Factor
      self.DomainSets = DomainSets
      self.nSlots = len(self.DomainSets)
      self.SubstOrder = self.nSlots // 2
      assert(self.nSlots % 2 == 0)


      if isinstance(SlotSymmetry, PermutationGroup):
         # permutation group object given directly
         self.SlotSymmetry = SlotSymmetry
      else:
         # SlotSymmetry gives a list of generators of slot symmetries
         self.SlotSymmetry = _make_group_with_minimal_bsgs(self.nSlots+2, SlotSymmetry)
      

      if LowerPerms is not None:
         for (LowerPerm,PermFactor) in LowerPerms:
            assert(self.SubstOrder == len(LowerPerm) and IsPermutation(LowerPerm))
      self.LowerPerms = LowerPerms
      self.ConnectionNeeded = not self.ThisIsLhs
      
   def __repr__(self):
      return "{}[{}]".format(self.Name, (2 * self.SubstOrder) * '.')
   
   def is_equal(self, other):
      if self.Name != other.Name:
         return False
      if self.nSlots != other.nSlots:
         return False
      if self.DomainSets != other.DomainSets:
         return False
      return True


def _MakeSlotSymmetry(OpTerms):
   """Make the symmetry group for permuting slots in the product described by
   OpTerms. Ts must already be ordered such that equal tensors come right next to
   each other."""

   # find total number of slots in the product
   nSlots = 0
   for Op in OpTerms:
      nSlots += Op.nSlots

   # there are two types of symmetries: The symmetries of the individual
   # tensors, moved to their respective places in the total contraction,
   # and the symmetries originating from the exchange of two equally named
   # tensors. Make generators for all of them.
   iSlot = 0

   Gens = []
   for (iT, T) in enumerate(OpTerms):
      # generators of intrinsic symmetries of tensor.
      if not T.ThisIsLhs:
         SymT = T.SlotSymmetry
      else:
         # this is an output tensor... cannot assume the symmetry of this.
         SymT = _make_group_with_minimal_bsgs(T.nSlots + 2, [])
      if isinstance(SymT, PermutationGroup):
         # permutation group object given---should, technically, already have
         # a minimal BSGS (as this is how we make them)
         SymT_Base, SymT_Gens = SymT.base, SymT.strong_gens
      else:
         # minimal BSGS given directly
         SymT_Base, SymT_Gens = SymT
      n = T.nSlots

      # move generators of intrinsic symmetry of the tensor
      for Gen_ in SymT_Gens:
         Gen = list(Gen_)
         assert(len(Gen) == n + 2)
         P = list(range(nSlots+2))   # + 2 for sign.
         P[iSlot:iSlot+n] = [o + iSlot for o in Gen[:n]]
         if Gen[-2] > Gen[-1]:
            P[-1], P[-2] = P[-2], P[-1]
            #P[n:n+2] = Gen[-2:]
         Gens.append(P)

      # check if this tensor can be exchanged with the last
      if iT != 0 and OpTerms[iT-1].is_equal(T):
         # yes... previous tensor had same name/type.
         # make generator for exchange of all slots of current tensor
         # with last tensor.
         P = list(range(nSlots+2))   # + 2 for sign.
         P = P[:iSlot-n] + P[iSlot:iSlot+n] + P[iSlot-n:iSlot] + P[iSlot+n:]
         Gens.append(P)
      iSlot += n
   SlotGroup = _make_group_with_minimal_bsgs(nSlots+2, Gens)
   return SlotGroup


def argsort(seq):
    # http://stackoverflow.com/questions/3071415/efficient-method-to-calculate-the-rank-vector-of-a-list-in-python
    return sorted(range(len(seq)), key=seq.__getitem__)


FCanEntry = namedtuple("FCanEntry", ["g","s"])

class FCanonicalizationContext(object):
   def __init__(self, nPermLength, SlotSym, DummyClasses):
      self.nPermLength = nPermLength # length of a permutation representing an index configuration
      self.SlotSym = SlotSym # permutation group of the slot symmetry
      self.DummyClasses = [] # list of consecutive dummy indices which can be permuted (i.e., indices into AllIndicesCan)
      for DummyClass in DummyClasses:
         if len(DummyClass) != 0:
            self.DummyClasses.append(DummyClass) # don't need empty ones.

      # this maps slot indices into a new order; namely, the order which
      # makes the entries of the slot symmetry group's base come first.
      # We will canoniaclize with respect to this order.
      self.af_slot_priority = tuple(SlotSym.base + [o for o in range(nPermLength) if o not in SlotSym.base])
      self.af_slot_priority_inv = _af_invert(self.af_slot_priority)

      # make some intermediate transformed transversals of the slot symmetry group
      # which are needed in the inv(g)-based canonicalization routines:
      #
      # namely, a list of inverses of array forms of (af_slot_priority_inv * p * af_slot_priority), where
      # p goes over the transversals of the slot symmetry bases
      self.StarredTransversals = []
      for (iBase, (BaseI, TraI)) in enumerate(zip(self.SlotSym.base, self.SlotSym.basic_transversals)):
         assert(self.af_slot_priority[iBase] == BaseI)

         p_star_list = []
         for p in TraI.values():
            #p_star = _af_invert(_af_rmul(_af_rmul(self.af_slot_priority_inv, p.array_form), self.af_slot_priority))
            p_star_list.append(self._af_p_star_trafo(p.array_form))
            p_star = self._af_p_star_trafo(p.array_form)
            for ix in range(iBase):
               # these things should be identity permutations for all indices
               # before the current base point (which is mapped to iBase under the star transformation)
               assert(p_star[ix] == ix)
         self.StarredTransversals.append(p_star_list)
   
   def _af_g_star_trafo(self, g):
      # compute g_star = (g * af_slot_priority)^{-1}
      return _af_invert(_af_rmul(g, self.af_slot_priority))

   def _af_g_star_trafo_inv(self, g_star):
      # compute inverse of _af_star_trafo:
      # g_star = (g * af_slot_priority)^{-1}
      # g_star^{-1} = g * af_slot_priority
      # g_star^{-1} af_slot_priority^{-1} = g
      # (af_slot_priority * g_star)^{-1}  = g
      return _af_invert(_af_rmul(self.af_slot_priority, g_star))
      #_g = _af_invert(_af_rmul(self.af_slot_priority, g_star))
      #assert(self._af_g_star_trafo(_g) == g_star)
      #return _g
      
   def _af_p_star_trafo(self, p):
      # compute p_star =  (af_slot_priority_inv * p * af_slot_priority)^{-1}
      return _af_invert(_af_rmul(_af_rmul(self.af_slot_priority_inv, p), self.af_slot_priority))

   def _af_p_star_trafo_inv(self, p_star):
      return _af_invert(_af_rmul(_af_rmul(self.af_slot_priority, p_star), self.af_slot_priority_inv))

   def _cmp_perm(self, a, b):
      """compare two permutations with respect to the canonicalization order"""
      # I think the actual order regarding which I canonicalize is not 'q < h'
      # but rather Order * q < Order * h. Otherwise I cannot directly use my BSGS
      # for the slot symmetries (note: I need to canonicalize with regard to *some*
      # order (and it may not depend on g), but which order does not really matter
      # as long as it is always the same one for equivalent sets of input tensors).
      #
      # ...in principle it might be a good idea to canonicalize with respect to
      #     InvOrder * g * Order
      # This would at least put identity first...
      return cmp(_af_rmul(a, self.af_slot_priority), _af_rmul(b, self.af_slot_priority))

   def _CanonicalizeIndexRenamesInverse(self, ginv_):
      # find the minium permutation of (d * g)
      # ...but input is inv(g) and output is inv(d*g) for technical reasons.
      
      # make a copy of the input permutation---don't change it in-place
      ginv = [o for o in ginv_]

      # go through all subsets of permutable sets of dummies.
      # (i.e., dummies belonging to the same index space)
      for DummyClass in self.DummyClasses:
         # sanity check. Dummy indices of a domain are supposed to
         # come in consecutive order and in pairs.
         assert(len(DummyClass) % 2 == 0)
         for i in range(len(DummyClass)):
            assert(DummyClass[i] == DummyClass[0]+i)
         iFirst = DummyClass[0]
         nPairs = len(DummyClass)//2
         iLast = iFirst + 2*nPairs
         assert(DummyClass[-1]+1 == iLast)
         
         # for each dummy pair (e.g., ii in iijjkkll), re-order 
         # in such a way that the lower slot index comes first.
         for iPair in range(nPairs):
            i0 = iFirst + 2*iPair
            i1 = i0 + 1
            if ginv[i0] > ginv[i1]:
               ginv[i0], ginv[i1] = ginv[i1], ginv[i0]

         # get first (=lower) slot index of every pair.
         ginv_i0s = ginv[iFirst:iLast:2]
         # sort them lexicographically
         iPairsSorted = argsort(ginv_i0s)
         assert(len(iPairsSorted) == nPairs)
         
         # ...and reorder all pairs accordingly
         ginv_dummies_new = []
         for iPair in iPairsSorted:
            i0 = iFirst + 2*iPair
            i1 = i0 + 1
            ginv_dummies_new.append(ginv[i0])
            ginv_dummies_new.append(ginv[i1])
         ginv = ginv[:iFirst] + ginv_dummies_new + ginv[iLast:]
      return tuple(ginv)



   @property
   def nSlots(self):
      return self.nPermLength - 2

   def CanonicalizeNaive(self, g):
      # make a group for the permutations representing index name exchanges.
      ExchSym = MakeDummyExchangeGroup(self.nSlots, self.DummyClasses)

      SlotPerms = self.SlotSym.generate(af=True)
      DummyPerms = list(ExchSym.generate(af=True))
      af_g = g.array_form

      h = tuple(af_g)
      for s in SlotPerms:
         for d in DummyPerms:
            q = tuple(_af_rmul(d, _af_rmul(af_g, s)))
            # ^- that is _[i] = d[g[s[i]]]
            # (required like this because we list for each slots i which index p[i]
            #  we take from the canonicalized index list)
            if self._cmp_perm(q, h) < 0:
               h = q

      # FIXME: doesn't check for vanishing due to anti-symmetry.
      return list(h)

   def CanonicalizeHalfNaive(self, g):
      # this one does the full iteration of the slot symmetries 's',
      # but finds the 'd' permutation implicitly in closed form via sorting.
      # To make this easy, the permutation order is decided based on g'^{-1},
      # rather than g itself (or some permutation of it).
      # (note: if g[i] denotes the index label which goes into slot 'i', then
      #  (g^{-1})[i] denotes into which slot the index label i goes.
      #  since we know which index labels belong to dummies, and which
      #  transformations are allowed there, the dummy parts of the permutation
      #  can easily be minimized lexicographically)
      SlotPerms = self.SlotSym.generate(af=True)

      af_g = g.array_form
      af_g_inv = _af_invert(af_g)

      min_g = tuple(_af_rmul(self.af_slot_priority_inv, af_g_inv))
      for s_inv in SlotPerms:
         # ^- note: this technically iterates over s, not s^{-1}.
         #    But since SlotPerms is a group, that gives the same
         #    results.
         q = tuple(self._CanonicalizeIndexRenamesInverse(_af_rmul(self.af_slot_priority_inv, _af_rmul(s_inv, af_g_inv))))
         if _af_invert(q) < _af_invert(min_g):
            min_g = q

      # FIXME:
      # - doesn't check for vanishing due to anti-symmetry.
      return list(_af_rmul(_af_invert(min_g), self.af_slot_priority_inv))


   def CanonicalizeRegular(self, g, AllowFreeIndexPerm=False):
      # list of remaining permutations we still have to consider.
      # In the g_star-form:
      #     g_star = (g * af_slot_priority)^{-1}
      # We also keep track of the 's' (in the s_star form) we needed to
      # get to the current candidate minimum d*g*s.
      dgps_star = [FCanEntry(self._af_g_star_trafo(g.array_form), list(range(g.size)))]
      #dgps_star = self._CanonicalizeIndexRenamesInverse(dgps_star[0].g)
      #return FCanEntry(self._af_g_star_trafo_inv(dgps_star), None)

      for (iBase, (BaseI, StarredTransverasl)) in enumerate(zip(self.SlotSym.base, self.StarredTransversals)):
         assert(self.af_slot_priority[iBase] == BaseI)

         valI_min = self.nPermLength
         next_dgps_star_candidates = []
         for (af_g_star, former_inv_s_star) in dgps_star:
            for p_star in StarredTransverasl:
               dgp_star = _af_rmul(p_star, af_g_star)
               dgp_star = self._CanonicalizeIndexRenamesInverse(dgp_star)

               valI = dgp_star.index(iBase) # <- that's _af_invert(dgp_star)[iBase] == dgp[BaseI]
               if valI < valI_min:
                  valI_min = valI
                  next_dgps_star_candidates = []
               if valI_min == valI:
                  next_dgps_star_candidates.append(FCanEntry(dgp_star, _af_rmul(p_star, former_inv_s_star)))

         # these are all the permutations and stuff we still have to consider as potentially canonical
         # (all of them gave minimum values on all previous base points)
         dgps_star = next_dgps_star_candidates

         # remove duplicates (note: required for anti-symmetry test).
         dgps_star = [FCanEntry(list(s),t) for (s,t) in (dict((tuple(u),v) for (u,v) in dgps_star).items())]
      dgps_star.sort(key = lambda x: _af_invert(x.g))
      # ^- this does the same as 'dgps.sort(lambda x,y: self._cmp_perm(x.g, y.g))' after the transformation.
      dgps = [FCanEntry(self._af_g_star_trafo_inv(dgp_star.g), self._af_p_star_trafo_inv(dgp_star.s)) for dgp_star in dgps_star]

      if len(dgps) > 1:
         # note: it's okay if we have more than one permutation here. The stuff
         # above really only fixes the baseI points of S. They might still differ
         # in the dummies.

         # In the case that we get two equivalent permutations which differ only
         # in sign (last two entries), we found that the tensor contraction vanishes
         # due to symmetry.
         dgp0, dgp1 = dgps[0:2]
         if dgp0.g[:self.nSlots] == dgp1.g[:self.nSlots]:
            return 0 # vanishing due to anti-symmetry
         assert(self._cmp_perm(dgps[0].g, dgps[1].g) != 0)
      return dgps[0]




FSlotInfo = namedtuple("FSlotInfo", ["iEop", "DomainSets", "OnLhs", "COpType"])
FPairedSlot = namedtuple("FPairedSlot", ["iSlotC", "iSlotD", "iEopC", "iEopD", "dsAllowed"])
FPairingInfo = namedtuple("FPairingInfo", ["IndexDs", "SlotPairings", "dsAll"])
FIndexDomainNamePair = namedtuple("FIndexDomainNamePair", ["ds", "Name"])

class FOpProd(object):
   def __init__(self, Factor, OpTerms):
      self.Ops = OpTerms
      self.Factor = Factor
      for Op in OpTerms:
         self.Factor *= Op.Factor
      
      self.iOpSt = [0]
      self.Slots = []
      for (iOp, Op) in enumerate(self.Ops):
         self.iOpSt.append(self.iOpSt[-1] + Op.nSlots)
         for iSlot in range(Op.nSlots):
            assert(Op.nSlots % 2 == 0)
            COpType = COP_Create if (iSlot < (Op.nSlots // 2)) else COP_Destroy
            self.Slots.append(FSlotInfo(iOp, Op.DomainSets[iSlot], Op.ThisIsLhs, COpType))
      self.nSlots = len(self.Slots)
      assert(self.nSlots == self.iOpSt[-1])
      
      # make lists of all slots connected to creation and destruction operators.
      self.iSlotCs = []
      self.iSlotDs = []
      self.iSlotsLhs = []
      self.iSlotsRhs = []
      for (iSlot, Slot) in enumerate(self.Slots):
         if Slot.COpType == COP_Create:
            self.iSlotCs.append(iSlot)
         elif Slot.COpType == COP_Destroy:
            self.iSlotDs.append(iSlot)
         else:
            assert(0)
         if Slot.OnLhs:
            self.iSlotsLhs.append(iSlot)
         else:
            self.iSlotsRhs.append(iSlot)
      
      self.PermsAndFactors = self._MakeOpPermsAndFactorsR(0, {tuple(range(self.nSlots)) : 1.0})
      
      self.iLhsOp = None
      self.iRhsOps = []
      for (iOp, Op) in enumerate(self.Ops):
         if Op.ThisIsLhs:
            assert(self.iLhsOp is None)
            self.iLhsOp = iOp
         else:
            self.iRhsOps.append(iOp)
      self._SetupSlotSymmetry()
      
   def _SetupSlotSymmetry(self):
      self.SlotSymmetry = _MakeSlotSymmetry(self.Ops)
      self.CanonicalizationContext = FCanonicalizationContext(self.nSlots + 2, self.SlotSymmetry, [list(range(self.nSlots))])
   
   def Canonicalize(self, g):
      gprime = self.CanonicalizationContext.CanonicalizeRegular(Permutation(g,size=self.CanonicalizationContext.nPermLength)).g
      #gprime = self.CanonicalizationContext.CanonicalizeHalfNaive(Permutation(g,size=self.CanonicalizationContext.nPermLength))
      assert(gprime is not None and gprime[-2] < gprime[-1]) # anti-symmetry not handled here.
      return gprime[:-2]

   
   def _MakeOpPermsAndFactorsR(self, iOp, PrevPermsAndFactors):
      if iOp >= len(self.Ops):
         return PrevPermsAndFactors
      Op = self.Ops[iOp]
      if not Op.LowerPerms:
         return self._MakeOpPermsAndFactorsR(iOp + 1, PrevPermsAndFactors)
      # transform Op's LowerPerms (which are written in terms of the Ops slots
      # alone---and only the lower ones at that) into permutations on the entire
      # operator product
      nSlotsLeft = self.iOpSt[iOp]
      nSlotsRight = self.iOpSt[iOp] + Op.nSlots // 2
      OpLower = list(range(nSlotsLeft, nSlotsRight))
      LowerPermsTranslated = {}
      for (LowerPerm, Factor) in Op.LowerPerms:
         p = tuple(list(range(nSlotsLeft)) + [OpLower[o] for o in LowerPerm] + list(range(nSlotsRight, self.nSlots)))
         LowerPermsTranslated[p] = Factor
      CombinedPermsAndFactors = {}
      for (PrevPerm, PrevFactor) in PrevPermsAndFactors.items():
         for (ThisPerm, ThisFactor) in LowerPermsTranslated.items():
            CombinedPermsAndFactors[tuple(_af_rmul(PrevPerm, ThisPerm))] = PrevFactor * ThisFactor
      return self._MakeOpPermsAndFactorsR(iOp+1, CombinedPermsAndFactors)

   def __str__(self):
      return "{:15.8f}  <{}>".format(self.Factor, " ".join(map(str,self.Ops)))

   def MakeDidxPermGroup(self):
      Gens = []
      for iididx in range(len(self.iSlotDs) - 1):
         ididx0 = self.iSlotDs[iididx]
         ididx1 = self.iSlotDs[iididx + 1]
         Gens.append(Permutation(ididx0, ididx1, size=self.nSlots))
      return _make_group_with_minimal_bsgs(self.nSlots, Gens)

      
   def FindPairedSlot(self, g1, iSlotD, AsEop):
      # find slot of cidx this didx is contracted to with the current 'g'
      i0 = g1[iSlotD]
      # ^- g1 maps slot indices to 'index indices' (which come in pairs which are contracted together)
      # find i1 as the index i0 is paired with
      if i0 % 2 == 0:
         i1 = i0 + 1
      else:
         i1 = i0 - 1
      # find the slot connecting to this paired index
      iSlotC = g1.index(i1)
      
      iEopC = self.Slots[iSlotC].iEop
      iEopD = self.Slots[iSlotD].iEop

      dsCommon = self.Slots[iSlotD].DomainSets.intersection(self.Slots[iSlotC].DomainSets)
      # ^- if this is empty, then slots are bound to non-overlapping domains -- cannot be contracted together.
      if AsEop:
         # treat slots as bound to tensorial E-operators; in this case the actually allowed
         # contraction domains also depend on the direction into which a contraction is
         # done (i.e., if it is a hole or a particle contraction).
         dsAllowed = set([])
         for ds in dsCommon:
            cicd = ds.CreatorIndexContractionDirections
            if (iEopD < iEopC and "L" in cicd) or (iEopD > iEopC and "R" in cicd) or (iEopD == iEopC and "C" in cicd):
               dsAllowed.add(ds)
         # ^- this may be empty even if dsCommon is not, if there is some domain overlap, but it is not compatible with the
         # actual contraction setting (e.g., trying to contract an external-orbital
         # creator to an external-orbital destructor to the right of it)
      else:
         # treat slots as belonging to raw numerical tensors
         dsAllowed = dsCommon
      return FPairedSlot(iSlotC, iSlotD, iEopC, iEopD, dsAllowed)
   
   def FindPairedSlots(self, g, AsEop = True):
      assert(len(g) == self.nSlots)
      IndexDs = self.nSlots * [None]
      SlotPairings = self.nSlots * [None]

      # for iSlotD in self.iSlotDs:
      iSlotsHandled = set([])
      def AddSlotPair(iSlotD):
         ps = self.FindPairedSlot(g, iSlotD, AsEop)
         iIdxC = g[ps.iSlotC]
         iIdxD = g[ps.iSlotD]
         iSlotsHandled.add(ps.iSlotC)
         iSlotsHandled.add(ps.iSlotD)
         
         assert(IndexDs[iIdxC] is None and IndexDs[iIdxD] is None)
         IndexDs[iIdxC] = ps.dsAllowed
         IndexDs[iIdxD] = ps.dsAllowed
         
         SlotPairings[ps.iSlotC] = (ps.iSlotD, ps)
         SlotPairings[ps.iSlotD] = (ps.iSlotC, ps)

      if AsEop:
         for iSlotD in self.iSlotDs:
            AddSlotPair(iSlotD)
      else:
         for iSlot1 in range(self.nSlots):
            if iSlot1 in iSlotsHandled:
               continue
            AddSlotPair(iSlot1)
         
         
      for iSlotD in range(self.nSlots):
         if iSlotD in iSlotsHandled:
            continue
      for dsSets in IndexDs:
         assert(dsSets is not None)
         assert(len(dsSets) in [0,1])
         # ^- the > 1 case would technically be allowed in some cases... but
         # not in closed-shell CC. Needs special handling and splitting of equations.
         # later.
      for dsSets in IndexDs:
         # no viable combination of indices to contract together
         if not dsSets:
            return None
      dsAll = set([])
      for dsSets in IndexDs:
         for ds in dsSets:
            dsAll.add(ds)
      return FPairingInfo(IndexDs, SlotPairings, dsAll)

   
   def MakeAdmissableSlotPairings(self):
      """return a dictionary mapping iSlotD to a list of all slot pairings for
      iSlotD which can possibly lead to a non-zero contraction.
      Note: This currently lists only didx slots as key and assumes you look
      for corresponding cidx slots."""
      SlotD_Pairings = {}
      # iterate over all slots bound to destruction operators
      for iSlotD in self.iSlotDs:
         PairableSlots = []
         # iterate over all slots bound to creation operators
         for iSlotC in self.iSlotCs:
            iEopC = self.Slots[iSlotC].iEop
            iEopD = self.Slots[iSlotD].iEop

            #print(":   iBase = {}  BaseI = {}  iSlotC = {}  iSlotC = {}  g = {}".format(iBase, BaseI, ididx, iSlotC, g1))
            dsCommon = self.Slots[iSlotD].DomainSets.intersection(self.Slots[iSlotC].DomainSets)
            # ^- if this is empty, then slots are bound to non-overlapping domains -- cannot be contracted together.
            dsAllowed = set([])
            for ds in dsCommon:
               cicd = ds.CreatorIndexContractionDirections
               if (iEopD < iEopC and "L" in cicd) or (iEopD > iEopC and "R" in cicd) or (iEopD == iEopC and "C" in cicd):
                  dsAllowed.add(ds)
            # ^- this may be empty even if dsCommon is not, if there is some domain overlap, but it is not compatible with the
            # actual contraction setting (e.g., trying to contract an external-orbital
            # creator to an external-orbital destructor to the right of it)
            if dsAllowed:
               PairableSlots.append(FPairedSlot(iSlotC, iSlotD, iEopC, iEopD, dsAllowed))
         SlotD_Pairings[iSlotD] = PairableSlots
      return SlotD_Pairings


   def MakeDidxPairingsList(self):
      # for each didx, find all cidx it can possibly be paired with, and determine
      # the corresponding index domains
      SlotD_Pairings = self.MakeAdmissableSlotPairings()

      # make all valid combinations of such pairings for all destruction slots
      # The output is a list of individual pairing lists; each inner list
      # thereby pairs up each upper index with a lower index.
      PairingsList = []
      N = len(self.iSlotDs)
      def MakePairingCombinationsR(iiSlotD, Pairings):
         if iiSlotD == N:
            # found one upper index for each lower index. that's all.
            PairingsList.append(Pairings)
         else:
            # go through each creator iu the current destruction index il can be paired with.
            iSlotD = self.iSlotDs[iiSlotD]
            iSlotCsUsed = [ps.iSlotC for ps in Pairings]
            for ps in SlotD_Pairings[iSlotD]:
               if ps.iSlotC in iSlotCsUsed:
                  continue # already used this one.
               MakePairingCombinationsR(iiSlotD + 1, Pairings + [ps])
      MakePairingCombinationsR(0, [])
      return PairingsList

   
   def _FormatPermRepRaw(self, g):
      """format a tensor contraction given by the permutation 'g' into an
      expression with explicit indices and index names"""
      assert(IsPermutation(g))
      assert(len(g) == self.nSlots)
      pi = self.FindPairedSlots(g, AsEop=False)
      
      IndexPool = FIndexPool(pi.dsAll)
      AssignedIndices = self.nSlots * [None]
      iSlotsHandled = set([])
      for iSlot0 in self.iSlotsLhs + self.iSlotsRhs:
         if iSlot0 in iSlotsHandled:
            continue
         (iSlot1, ps) = pi.SlotPairings[iSlot0]
         
         iIdx0 = g[iSlot0]
         iIdx1 = g[iSlot1]
         assert(abs(iIdx0 - iIdx1) == 1) # should be right next to each other
         assert(min(iIdx0,iIdx1) % 2 == 0) # and start at an even number.
         dsSets = ps.dsAllowed
         assert(len(dsSets) == 1) # see above
         ds = list(dsSets)[0]
         AsFree = iSlot0 in self.iSlotsLhs
         if iSlot1 in self.iSlotsLhs:
            assert(iSlot0 in self.iSlotsLhs)
            AsFree = True
         IndexName = IndexPool.Take(ds, AsFree)
         assert(AssignedIndices[iIdx0] is None and AssignedIndices[iIdx1] is None)
         # first by domain, then by free-ness (this prefers putting dummies to the left)
         prio = (ds.SortPriority,int(AsFree))
         #prio = ds.SortPriority
         AssignedIndices[iIdx0] = FIndexDomainNamePair(prio, IndexName)
         AssignedIndices[iIdx1] = FIndexDomainNamePair(prio, IndexName)
         iSlotsHandled.add(iSlot0)
         iSlotsHandled.add(iSlot1)
      for (iIndex, IndexName) in enumerate(AssignedIndices):
         assert(IndexName is not None)

      BoundOps = []
      for (iOp, Op) in enumerate(self.Ops):
         g_op = g[self.iOpSt[iOp] : self.iOpSt[iOp+1]]
         OpIndices = [AssignedIndices[o] for o in g_op]
         OpIndexNames = [o.Name for o in OpIndices]

         SortTensorInstanciationsByDomains = True
         if SortTensorInstanciationsByDomains:
            Exact = True
            if Exact:
               def PackKey(OpIndices):
                  return zip(*OpIndices)
               def UnpackKey(OpIndicesKey):
                  return [FIndexDomainNamePair(*o) for o in zip(*OpIndicesKey)]
            else:
               def PackKey(OpIndices):
                  return OpIndices
               def UnpackKey(OpIndicesKey):
                  return OpIndicesKey
               
            if iOp != self.iLhsOp:
               # sort indices by domains using op's slot permutations.
               # this is meant to effect that each unique tensor slice occurs
               # only once (e.g., mapping all f:ec and f:ce to one of the two).
               # Note: This step is optional. srci will happily make multiple views
               # into tensors if needed. The rest of the code works too without it.
               # This is mainly a cosmetical issue.
               Exact = False
               OpIndicesMinKey_ = None
               IdN = list(range(len(OpIndices)))
               for pOpSlot in Op.SlotSymmetry.generate():
                  OpIndicesPermutedKey = PackKey([OpIndices[pOpSlot(o)] for o in IdN])
                  if OpIndicesMinKey_ is None or OpIndicesPermutedKey < OpIndicesMinKey_:
                     OpIndicesMinKey_ = OpIndicesPermutedKey
               OpIndices = UnpackKey(OpIndicesMinKey_)
               # hmhm.. theoretically this should be good. But sorting by
               # FIndexDomainNamePair(ds.SortPriority, IndexName) directly,
               # in full lexicographical order, appeared to work MUCH better on tempest.
               # ...despite making some redundant tensors. Not sure what to make of that.
         
         OpIndexNames = [o.Name for o in OpIndices]
         Upper = OpIndexNames[:Op.nSlots//2]
         Lower = OpIndexNames[Op.nSlots//2:]
         if Op.SwitchUpperLower:
            Upper, Lower = Lower, Upper
         BoundOps.append(FIndexedTensor(Op.Name, Upper + Lower))
      return BoundOps


   def ExpandLowerPerms(self, g, OuterFactor):
      assert(len(g) == self.nSlots)
      assert(IsPermutation(g))
      Out = []
      for (SlotPerm, PermFactor) in self.PermsAndFactors.items():
         assert(IsPermutation(SlotPerm))
         gprime = _af_rmul(g, SlotPerm)
         Out.append((PermFactor * OuterFactor * self.Factor, gprime))
      return Out


   def IsConnected(self, g):
      """return whether a contraction is connected.
      Specifically, check whether Hs and Ts can be split into two distinct groups
      such that neither shares a contraction with the other.
      Parameters:
         iEopForSlot[iSlot]: map slot index iSlot in tensor product to
            iEop from which it comes
         ConnectionNeeded[iEop]: array of bool; if ConnectionNeeded[iEop] == True, iEop
            is assumed to either be a T or a H which needs connectivity.

      Notes:
         - In particular, Lambdas need not be connected
         - Operators with ConnectionNeeded[iOp] == False are ignored.
         (also indirect connections over such operators are not traced)
         - Scalar operators cannot possibly be connected and are ignored."""
      iVisited = set()
      SlotsFrom = {}
      SlotsTo = {}
      def Dfs(iEop):
         # depth first search
         if iEop in iVisited or not self.Ops[iEop].ConnectionNeeded:
            return
         iVisited.add(iEop)
         
         for iSlot in range(self.iOpSt[iEop], self.iOpSt[iEop+1]):
            Dfs(self.Slots[SlotsFrom[iSlot]].iEop)
            Dfs(self.Slots[SlotsTo[iSlot]].iEop)
      # find one of the eops which need to be connected, and check if all others
      # which need to be connected can be reached by tracing its indices.
      nOpsToVisit = 0
      iSomeEopToConnect = -1
      for (iEop, Op) in enumerate(self.Ops):
         if not Op.ConnectionNeeded:
            continue
         if Op.nSlots == 0:
            # scalar ops can never be connected over indices, since they do not have any.
            continue
         iSomeEopToConnect = iEop
         nOpsToVisit += 1
      if iSomeEopToConnect != -1:
         pi = self.FindPairedSlots(g)
         for (iSlot0, (iSlot1, _)) in enumerate(pi.SlotPairings):
            SlotsFrom[iSlot0] = iSlot1
            SlotsTo[iSlot1] = iSlot0
         Dfs(iSomeEopToConnect)
      return len(iVisited) == nOpsToVisit


def GetDidxWeightAndCharacter(DidxPerm):
   """Determine a permutation's contraction character and prefactor:
      Rules for slater-determinant particle-hole theorems:
      - particle contractions: upper right to lower left
      - hole contractions: lower right to upper left
      - each hole contraction gives a factor of (-1).
      - each closed loop gives a factor of (-ORB_OCC)."""
   nCycles = len(FindCycles(DidxPerm))
   nHoleCo = 0
   for iu,il in enumerate(DidxPerm):
      if il > iu:
         nHoleCo += 1
   Prefactor = (-1)**nHoleCo * (-ORB_OCC)**nCycles
   return Prefactor


def FmtIndexList1(L):
   return "".join(L)

def FmtIndexedTensor(o):
   return "{}[{}]".format(o.Name, FmtIndexList1(o.Indices))


def ProcessOpProd(OutputEqs, OpProd, OuterFactor, ConnectedOnly):
   t0 = time()
   #DidxPermGroup = OpProd.MakeDidxPermGroup()
   
   OutputPerms = {}
   nPermActual = [0,0]
   LhsOps = [None]

   def MakePermForPairingList(pl):
      ginv = []
      assert(len(OpProd.iSlotCs) == len(OpProd.iSlotDs))
      for pi in pl:
         ginv.append(pi.iSlotC)
         ginv.append(pi.iSlotD)
      return _af_invert(ginv)

   def ProcessPairing(pl):
      g = MakePermForPairingList(pl)

      if ConnectedOnly and not OpProd.IsConnected(g):
         return
      
      pi = OpProd.FindPairedSlots(g)
      if pi is None:
         # some pairings implied by g were invalid.
         return

      # compute permutation DidxPerm which aligns the creation indices of eops
      # to the corresponding destruction indices of eops.
      assert(OpProd.nSlots % 2 == 0)
      DidxPerm = (OpProd.nSlots//2) * [None]
      for (iSlot0, (iSlot1, PairedSlot)) in enumerate(pi.SlotPairings):
         assert(isinstance(PairedSlot, FPairedSlot))
         DidxPerm[OpProd.iSlotCs.index(PairedSlot.iSlotC)] = OpProd.iSlotDs.index(PairedSlot.iSlotD)
      #DidxPerm = [OpProd.iSlotDs.index(d[iSlotD]) for iSlotD in OpProd.iSlotDs]
      #DidxPerm = [OpProd.iSlotDs.index(d.index(iSlotD)) for iSlotD in OpProd.iSlotDs]
      assert(IsPermutation(DidxPerm))
      # ^- all three variants give the same result. First one should be the
      # safest, in principle.
      Factor = OuterFactor * GetDidxWeightAndCharacter(DidxPerm)
      
      # assemble new terms generated by the current destructor index permutation.
      TermList = OpProd.ExpandLowerPerms(g, Factor)
      for (Factor, gprime) in TermList:
         gprime_ = tuple(gprime)
         OutputPerms[gprime_] = OutputPerms.get(gprime_, 0.) + Factor
      
      nPermActual[0] += 1
      pass
   
   N = OpProd.nSlots
   PairingCombs = OpProd.MakeDidxPairingsList()
   t0 = time()
   for pl in PairingCombs:
      ProcessPairing(pl)
   

   CanonicalPerms = {}
   for (g,v) in sorted(OutputPerms.items()):
      if abs(v) < 1e-10:
         continue
      gprime_ = tuple(OpProd.Canonicalize(g))
      CanonicalPerms[gprime_] = CanonicalPerms.get(gprime_, 0.) + v

   t1 = time()
   nDidxPerm = nPermActual[0]
   if nDidxPerm != 0:
      print("{:<75} -> nPerm = {:<8}   nEqs = {:<8}   (time: {:.2f} sec)".format(str(OpProd), nDidxPerm, len(CanonicalPerms), t1-t0))

   CanonicalTerms = {}
   for (gprime, Factor) in sorted(CanonicalPerms.items()):
      BoundOps = OpProd._FormatPermRepRaw(gprime)
      LhsOp = BoundOps[OpProd.iLhsOp]
      if LhsOps[0] is None:
         LhsOps[0] = LhsOp
      else:
         assert(LhsOps[0] == LhsOp)
      RhsOps = [BoundOps[o] for o in OpProd.iRhsOps] 
      F, Ts = 1., RhsOps
      T = " ".join(FmtIndexedTensor(o) for o in Ts)
      if F != 0:
         CanonicalTerms[T] = CanonicalTerms.get(T, 0.) + F * Factor


   nLines = 0
   LhsOp = None
   if LhsOps[0] is not None:
      # ^- may be None if there are *no* resulting equations.
      LhsOp = LhsOps[0]
   for (k,v) in sorted(CanonicalTerms.items()):
      if abs(v) < 1e-10:
         continue
      sLhs = "{}[{}]".format(LhsOp.Name, "".join(LhsOp.Indices))
      if 0:
         Eq = "{} += {} {}".format(sLhs, FloatToStr1(v), k)
         print("      {}".format(Eq))
      key = (sLhs, k)
      OutputEqs[key] = OutputEqs.get(key, 0.) + v
   pass



def WriteEquationList(FileNameOut, NewEqs):
   with open(FileNameOut,"w") as File:
      for ((Lhs,T), Factor) in sorted(NewEqs.items()):
         if abs(Factor) < 1e-8:
            continue
         Line = "{} += {} {}\n".format(Lhs,FloatToStr1(Factor),T)
         SkipRedundant = True
         if SkipRedundant:
            # See paper: This skips the redundant equations arising from the 0-space
            # of the permutational overlap matrix.
            PrintWarning = False
            if Lhs.startswith("r3[ABCIJK]") and any([abs(abs(Factor) - o) < 1e-6 for o in [.1,.2,.4,.8]]):
               if PrintWarning: print("!WARNING: skipped %s" % (Line[:-1]))
               continue
            if Lhs.startswith("r4[ABCDIJKL]") and any([abs(abs(Factor) - o) < 1e-6 for o in [1./14.,2./14.,4./14.,8./14.,16/14.]]):
               if PrintWarning: print("!WARNING: skipped %s" % (Line[:-1]))
               continue
            if 1:
               for rt in ["r","t"]:
                  for i in range(10):
                     Line = Line.replace("%s%i["%(rt,i), "%s[" % rt)
         if _ThisIsPython2:
            File.write(Line.encode("utf-8"))
         else:
            File.write(Line)
      pass
   pass
   print("* wrote '{}'".format(FileNameOut))


def MakeExchPerm(n):
   """Generates a tensor symmetry specification representing slot permutation symmetries
   like closed-shell T-amplitudes have (e.g., T[abij] = T[baji], and their heigher order
   generalizations). 'n' denotes the number of indices.
   """
   def MakeP(n, L):
      Gens = [Permutation(tuple(o), size=n+2) for o in L]
      if not Gens:
         Gens = [Permutation(size=n+2)]
      return Gens

   assert(n%2 == 0)
   m = n//2
   def DP(p):
      return tuple([o for o in p] + [o + m for o in p] + [2*m,2*m+1])
   Exch01 = DP([1,0] + list(range(2,m)))
   Cycle = DP([m-1] + list(range(m-1)))
   if n <= 2:
      return MakeP(n, [])
   elif n == 4:
      return MakeP(n, [Cycle])
   return MakeP(n, [Cycle,Exch01])


class FCiMethodDesc(object):
   def __init__(self, MethodName, OrdersT, LinearOnly, ConnectedOnly, MaxTotalOrder=None):
      self.MethodName = MethodName
      self.OrdersT = OrdersT
      self.LinearOnly = LinearOnly
      self.ConnectedOnly = ConnectedOnly
      self.MaxTotalOrder = MaxTotalOrder      


def MakeEqSet(OutputFile, MethodDesc, EqDecl, OuterFactor):
   import re

   dsClo = set([FIndexDomain("clo", 'c', 'ijklmnop', "R", SortPriority=20)])
   dsExt = set([FIndexDomain("ext", 'e', 'abcdefgh', "L", SortPriority=10)])
   dsUni = dsClo | dsExt
   
   print("\n *** PROCESSING EQUATION SET {}: {} -> {}\n".format(MethodDesc.MethodName, EqDecl, OutputFile))

   def MakeOpTerms(TermDecl_, Mods = []):
      """This function pulls apart declarations of tensor operators in the equation specification,
      and makes FOpTerm lists out of them.
      For example, it generates lists representing two FOpTerms for f and W for the tensor 'H' (Hamiltonian),
      or more complex lists for ExprT which represent the cluster operator.
      """
      TermDecl = TermDecl_.strip()
      # check if a modifier is applied to the current tensor term. E.g., "lhs(co(ExprT))".
      # in this case we would pull out the 'lhs(...)' here and recurse on 'co(ExprT)'.
      # This way we just collect all modifier expressions until we reached an actual tensor
      # declaration. See below for meaning of modifiers.
      m = re.match(r"(\w+)\((.*)\)", TermDecl)
      if m:
         # TermDecl matches Modifier(...subexpr)
         # This means that some sort of modifier is applied to it.
         Modifier = m.group(1)
         SubExpr = m.group(2)
         return MakeOpTerms(SubExpr, [Modifier] + Mods)

      # apply T-op expression to the term? (i.e., exp(T) in coupled-cluster case, but only T in CI case)
      ExpTerm = False
      # this term is the one which goes out at the end? (i.e., on the left hand side of the equation)
      IsLhs = False
      # use GetLeftPermsAndFactors with the permutation factors for residuals?
      ApplyContraVariantPermutations = False
      # conjugate; i.e., switch lower and upper indices in the E-op?
      Adjoint = False
      # skip 0'th order term? (constant 1)
      SkipOrder0 = False
      # Use only orders in T up to 1 in ExprT
      MaxOrderT = None
      
      for Mod in Mods:
         if Mod == "lhs":
            IsLhs = True
         elif Mod == "adj":
            Adjoint = True
         elif Mod == "co":
            ApplyContraVariantPermutations = True
         elif Mod == "no_order_0":
            SkipOrder0 = True
         elif Mod == "max_order_1":
            MaxOrderT = 1
         else:
            raise Exception("Modifier '{}' in CC term declaration '{}' not recognized".format(Mod, EqDecl))
      CoOrContra = "contra" if ApplyContraVariantPermutations else "co"

      TensorName = TermDecl
      if TensorName == "H":
         # Make Hamiltonian terms: Fock matrix (f) and 2-electron integrals (W).
         TermsH = []
         UseOnlyExchangeSymForH = False
         if UseOnlyExchangeSymForH:
            # use only 2-fold symmetry on W and no symmetry on f. This actually
            # is sufficient to get the minimum number of equations, and does not involve
            # any operations which would require complex conjugation in the complex case.
            TermsH += [(1., [FOpTerm("W", 0.5, [dsUni, dsUni, dsUni, dsUni], [[1,0,3,2]])])]
            TermsH += [(1., [FOpTerm("f", 1.0, [dsUni, dsUni], [])])]
         else:
            # use full 8-fold symmetry on W and 2-fold symmetry on f
            # Note: W is given in exchange-order:  W[abcd] = <ab|1/r|cd> = (ac|bd)  (the latter being Mulliken order)
            TermsH += [(1., [FOpTerm("W", 0.5, [dsUni, dsUni, dsUni, dsUni], [[2,1,0,3], [0,3,2,1], [1,0,3,2]])])]
            TermsH += [(1., [FOpTerm("f", 1.0, [dsUni, dsUni], [[1,0]])])]
         if len(Mods) != 0:
            raise Exception("Modifiers '{}' in CC term declaration '{}' not allowed for Hamiltonian terms".format(repr(Mods), EqDecl))
         return TermsH
      elif TensorName in ["r", "t", "l"]:
         # straight residual, amplitude, or lagrange multiplier
         TermsRT = []
         for OrderRT in (0,) + MethodDesc.OrdersT:
            if SkipOrder0 and OrderRT == 0:
               continue
            if Adjoint:
               ds = OrderRT * [dsClo] + OrderRT * [dsExt]
            else:
               ds = OrderRT * [dsExt] + OrderRT * [dsClo]
            TermsRT += [(1., [FOpTerm("{}{}".format(TensorName, OrderRT), 1.0, ds, MakeExchPerm(2*OrderRT),
               SwitchUpperLower = Adjoint, ThisIsLhs = IsLhs,
               LowerPerms = GetLeftPermsAndFactors(OrderRT, CoOrContra))])]
         return TermsRT
      elif TensorName in "ExprT":
         # method-defining expression applied to T, for CC methods that is exp(T), for CI methods just T itself.
         # Make the corresponding terms.
         if len([Mod for Mod in Mods if not Mod.startswith("max_order_")]) != 0:
            raise Exception("Modifiers '{}' in CC term declaration '{}' not allowed for expr(T) terms".format(repr(Mods), EqDecl))

         ExprT = MakeExpTOpDecl(2+max(MethodDesc.OrdersT), MethodDesc.OrdersT, Commutative=True, LinearOnly=MethodDesc.LinearOnly, MaxOrderT=MaxOrderT)
         TOps = []
         for OrderT in range(max(MethodDesc.OrdersT)+1):
            TOps += [FOpTerm("t{}".format(OrderT), 1.0/_Factorial(OrderT), OrderT * [dsExt] + OrderT * [dsClo], MakeExchPerm(2*OrderT))]

         TermsExprT = []
         for (DeclT, FactorT) in sorted(ExprT.items()):
            TermsT = []
            for it in DeclT:
               TermsT += [TOps[it]]
            TermsExprT.append((FactorT, TermsT))
         return TermsExprT
      elif TensorName in ["Nrm", "EHy", "ECorr"]:
         # these are scalar tensors.
         if Adjoint or ApplyContraVariantPermutations:
            raise Exception("adj() and co() are not valid modifiers for scalar tensor '{}' in eq '{}'".format(TensorName, EqDecl))
         if TensorName == "ECorr":
            TensorName = "r"  # r[0] is where the correlation energy sits.
            
         ScalarTerm = (1., [FOpTerm(TensorName, 1.0, [], [],
               SwitchUpperLower = Adjoint, ThisIsLhs = IsLhs,
               LowerPerms = GetLeftPermsAndFactors(0, CoOrContra))])
         return [ScalarTerm]
      else:
         raise Exception("Tensor operator name '{}' in '{}'not recognized.".format(TensorName, EqDecl))
      pass
         
   # parse and process the input equation type declaration.
   assert(EqDecl[0]=='<' and EqDecl[-1] == ">")
   TermDecls = EqDecl[1:-1].strip().split(",")
   EqParts = []
   for TermDecl in TermDecls:
      EqParts.append(MakeOpTerms(TermDecl))
   
   OpProds = []
   # unpack all combinations of input terms into individual FOpProd objects
   def MakeOpProdsR(PrevTerms, PrevFactor, RemainingEqParts):
      if RemainingEqParts:
         # recurse
         EqPart = RemainingEqParts[0]
         for (OpFactor, OpTerms) in EqPart:
            MakeOpProdsR(PrevTerms + OpTerms, PrevFactor * OpFactor, RemainingEqParts[1:])
      else:
         # reached end of recursion. Add term to output operator products.
         # note: should check max order and stuff here, to see if the expr
         # actually can make a non-zero contribution and/or should actually be included.
         TermsAll = PrevTerms
         OpProds.append(FOpProd(PrevFactor, TermsAll))
   MakeOpProdsR([], OuterFactor, EqParts)
         
   OutputEqs = {}
   for OpProd in OpProds:
      ProcessOpProd(OutputEqs, OpProd, 1., ConnectedOnly = MethodDesc.ConnectedOnly)
      
   print("\n  Number of output equations: {} (including redundant)".format(len(OutputEqs)))
   WriteEquationList(OutputFile, OutputEqs)
   print()
      


def MakeCcSet(BaseFileName, MethodDesc):
   # make output directory if it does not already exist
   import os
   try:
      os.mkdir("eqs")
   except OSError:
      pass
   
   # collect actual equation lists needed for the sub-algorithms
   # required in the CC(n) and CI(n) methods
   OutputFileTemplate = "eqs/{}-{{}}.itfaa".format(BaseFileName)
   ItfaaFiles = []
   def MakeEqs(AlgoName, EqDecl, Factor):
      OutputFile = OutputFileTemplate.format(AlgoName)
      MakeEqSet(OutputFile, MethodDesc, EqDecl, Factor)
      ItfaaFiles.append((AlgoName, OutputFile))
   
   # algo incrementing residuals by a multiple of the amplitudes (needed
   # as part of the CI(n) residual equations in addition to the main part
   # of the residual, see srci source code)
   MakeEqs("A2r", "<lhs(co(adj(r))), ECorr, no_order_0(t)>", -1.0)
   # algo for incrementing amplitudes by residuals. This is used as the
   # actual amplitude update in the CC(n) and CI(n) solvers, in order
   # to realize the perturbative update
   #
   #    t[ABC..IJK...] -= r[ABC...IJK...] / (eps[A] + eps[B] + ... - eps[I] - eps[J] -...)
   #
   # The input residuals to this algorithm have the denominators scaling
   # already applied.
   MakeEqs("R2a", "<lhs(co(adj(t))), no_order_0(r)>", -1.0)
   # algo to compute approximate squared norm of expr(T)|Phi0> vector. For CI,
   # this computes the actual squared norm, for CC, only a non-hermitian approximation
   MakeEqs("Nrm", "<lhs(Nrm), adj(t), max_order_1(ExprT)>", 1.0)
   #   MakeEqs("Nrm", "<lhs(Nrm), adj(t), ExprT>", 1.0)
   # algo to compute Hyleraas functional estimate of correlation energy
   # (converges a bit faster. For CEPA(0), this computes the actual Hyleraas
   # functional, not only an approximately 2nd order scaling term)
   MakeEqs("EHy2", "<lhs(EHy), adj(r), max_order_1(ExprT)>", 1.0)
   #   MakeEqs("EHy2", "<lhs(EHy), adj(r), ExprT>", 1.0)
   # algo for evaluating the main part of the residual
   MakeEqs("Res", "<lhs(co(adj(r))), H, ExprT>", +1.0)

   def ExportToSrci():
      from os import path
      import itfaa_to_srci
      SrciPath = "srci_inline_files"
      iExportArg = sys.argv.index("--export")
      if len(sys.argv) > iExportArg + 1:
         SrciPath = sys.argv[iExportArg+1]
      
      FileName_SrciInl = path.join(SrciPath, "{}-R.inl".format(BaseFileName))
      NamespaceName = BaseFileName + "R"
      itfaa_to_srci.ConvertFiles(FileName_SrciInl,
         NamespaceName, MethodDesc.MethodName, ItfaaFiles)
      
   if "--export" in sys.argv:
      ExportToSrci()

if __name__ == "__main__":
   MakeCcSet("EqCisd", FCiMethodDesc("CISD", (1,2), LinearOnly=True, MaxTotalOrder=6, ConnectedOnly=False))
   MakeCcSet("EqCcsd", FCiMethodDesc("CCSD", (1,2), LinearOnly=False, ConnectedOnly=True))
   MakeCcSet("EqCisdt", FCiMethodDesc("CISDT", (1,2,3), LinearOnly=True, ConnectedOnly=False))
   MakeCcSet("EqCcsdt", FCiMethodDesc("CCSDT", (1,2,3), LinearOnly=False, ConnectedOnly=True))
   #MakeCcSet("EqCisdtq", FCiMethodDesc("CISDTQ", (1,2,3,4), LinearOnly=True, ConnectedOnly=False))
   #MakeCcSet("EqCcsdtq", FCiMethodDesc("CCSDTQ", (1,2,3,4), LinearOnly=False, ConnectedOnly=True))

