# Copyright (c) 2013-2019 Gerald Knizia
# 
# This file is part of the 'cc_eqs_pv' program
# (See https://sites.psu.edu/knizia/)
# 
# cc_eqs_pv is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# cc_eqs_pv is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with cc_eqs_pv (LICENSE_GPL_v3.txt). If not, see
# http://www.gnu.org/licenses/

"""This is an auxiliary file which converts lists of equations generated
by the main program into inline files for the srci C++ program (the Eq*-R.inl
files). Concretely, it collects the sets of used tensors and their usage
classes from the input itfaa equation sets, assigns them numbers, and
encodes all the actual contraction equations into the low level format
used by srci."""
from __future__ import print_function
import re
from collections import namedtuple

from os import path
import CodeGen
from CodeGen import Emit, Ind, UnInd, FloatToStr, BeginBlock, EndBlock

import platform
_ThisIsPython2 = (int(platform.python_version_tuple()[0]) == 2)
if _ThisIsPython2:
   def utf8_encode(s): return s.encode("utf8")
   def utf8_decode(s): return s.decode("utf8")
else:
   def utf8_encode(s): return s
   def utf8_decode(s): return s
   

def IndexDomain(c,UnivAsActive=False):
   if c.lower() in u"abcdefgh": return u"e" # external
   if c.lower() in u"ijklmnop": return u"c" # closed
   if c.lower() in u"vwxyzq": return u"a"
   if c.lower() in u"rstu": return u"u" if not UnivAsActive else u"a"
   assert(0)
   
def IndexSpin(c):
   return "_"

def IndexToBase(c):
   # convert an index into its base form; e.g., convert an
   # italic 'a' into an ascii 'a' (used to mark alpha spin)
   return c


FAsmError = Exception
FIndexedTensor = namedtuple("FIndexedTensor", ["Name", "Indices"])

def ReadIndexList(Text):
   # rules:
   #    - if the index list contains a comma, indices are separated
   #      by commas.
   #    - if it doesn't, each character is an index
   #    - except for groups in {} which form compound indices
   Indices = []
   FIndex = lambda s: s
   if u',' in Text:
      return map(FIndex,Text.split(u","))
   i = 0
   while i < len(Text):
      if Text[i] == u"{":
         # find end of index name
         i += 1
         iEnd = i + 1
         while iEnd < len(Text) and Text[iEnd] != u"}":
            iEnd += 1
         assert(Text[iEnd] == u"}")
         Indices.append(FIndex(Text[i:iEnd]))
         i = iEnd
      else:
         Indices.append(FIndex(Text[i:i+1]))
      i += 1
   return Indices

#TensorRe = re.compile(ur"([a-zA-Z_][a-zA-Z0-9:_{}\+-]*)(\[[a-zA-Z0-9{},]*\])?")
TensorRe = re.compile(r"([a-zA-Z_][a-zA-Z0-9:_{}\+-]*)(\[.*\])?")
# ^- that's for dealing with unicode indices...

def ReadIndexedTensor(Text):
   """reads occurence of indexed tensors in the text."""
   if _ThisIsPython2:
      assert(isinstance(Text,unicode))
   Match = TensorRe.match(Text)
   if ( Match == None ):
      raise FAsmError(u"did not understand tensor '{}'".format(Text))
   Groups = Match.groups()
   Name = Groups[0]
   # scalar tensors may ommit the index list.
   if Groups[1] != None:
      Indices = ReadIndexList(Groups[1][1:-1])
   else:
      Indices = []
   return FIndexedTensor(Name,Indices)


class FContraction(object):
   def __init__(self, Line, Factor=None, Lhs=None, Rhs=None):
      if Line is not None:
         try:
            Line = Line.strip()
            if Line.startswith("."):
               Line = Line[1:].strip()
            
            Factor = 1.
            ls = Line.split("+=")
            if len(ls) != 2:
               ls = Line.split("-=")
            if len(ls) != 2:
               raise Exception("Expected line containing += or -=")

            Lhs = ReadIndexedTensor(ls[0])
            Rhs = ls[1].split()
            try:
               # first entry on rhs may be a scalar prefactor. try to
               # convert it as such and remove from the list if successful.
               f = float(Rhs[0])
               Factor *= f
               del Rhs[0]
            except:
               pass
            # convert to tensor objects.
            Rhs = [ReadIndexedTensor(o) for o in Rhs]
         except Exception as e:
            raise Exception("While reading equation: '{}'\n'{}'".format(Line,e))
      else:
         assert(Lhs is not None and isinstance(Rhs,list) and isinstance(Factor,float))

      self.Factor = Factor
      self.Rhs = Rhs
      self.Lhs = Lhs
   def __unicode__(self):
      from CodeGen import FloatToStr
      def Fmt(T):
         return u"{}[{}]".format(T[0],"".join(T[1]))
      f = self.Factor
      Op = "+="
      if f < 0:
         f *= -1
         Op = "-="
      s = u"{} {} {} {}".format(Fmt(self.Lhs), Op, FloatToStr(f), " ".join(Fmt(o) for o in self.Rhs))
      return s
   def __str__(self):
      return utf8_encode(self.__unicode__())


def ReadEquationList(FileName):
   with open(FileName,"rb") as File:
      Lines = utf_decode(File.read()).splitlines()
   def StripComments(Line):
      ix = Line.find("//")
      if ix != -1:
         Line = Line[:ix]
      return Line.strip()
   # strip comments and empty lines
   Lines = [StripComments(Line) for Line in Lines]
   Lines = [Line for Line in Lines if Line != u""]
   Eqs = [FContraction(Line) for Line in Lines]
   return Eqs




def ConvertToSrci(MethodName, EqSets, MethodSpinClass):
   del CodeGen._OutputLines[:]
   def MakeCoStr(Eq, IndexRemap=None):
      def idx(x):
         if IndexRemap is not None:
            return "".join(IndexRemap(o) for o in x[1])
         else:
            return "".join(x[1])
      return ",".join(idx(o) for o in [Eq.Lhs] + Eq.Rhs)

   AllCos = sum([Cos for (SetName,Cos) in EqSets],[])
   AllTensors = set([])
   def ToNameAndDomain(T):
      Name,Indices = T
      Spins = [IndexSpin(o) for o in Indices]
      Domain = [IndexDomain(IndexToBase(o)) for o in Indices]
      return "{}:{}:{}".format(Name,"".join(Domain),"".join(Spins))
   for Co in AllCos:
      AllTensors |= set(map(ToNameAndDomain,[Co.Lhs]))
      AllTensors |= set(map(ToNameAndDomain,Co.Rhs))
   def CmpTensors_Key(A):
      # put residuals first, then amplitudes, then everything else.
      i0 = 7
      if A.startswith("r:"):
         i0 = 0
      elif A.startswith("t:"):
         i0 = 1
      elif A.startswith("W:") or A.startswith("f:"):
         i0 = 7
      else:
         i0 = 2
      return (i0, A)

   TensorIndices = {}
   TensorNames = {}
   for (iTensor, Tensor) in enumerate(sorted(AllTensors,key=CmpTensors_Key)):
      TensorIndices[Tensor] = iTensor
      TensorNames[iTensor] = Tensor

#struct FTensorDecl
#{
   #char const
      #*pName,
      #*pDomain, // name:domain. E.g., 't:eecc' is a T^ij_ab T2 tensor.
      #*pSymmetry; // ITF-style permutation symmetry declaration.
   #FTensorUsage
      #Usage;
#};
   def GetUsageClass(Name):
      if Name == "t": return "USAGE_Amplitude"
      elif Name == "r": return "USAGE_Residual"
      elif Name in "f W W_AA W_AB W_BB f_A f_B".split(): return "USAGE_Hamiltonian"
      elif Name.lower().startswith("rdm"): return "USAGE_Density"
      elif Name.lower().startswith("nrm"): return "USAGE_Density"
      elif Name.lower().startswith("ehy"): return "USAGE_Intermediate"
      else:
         if Name.startswith("I") or Name.startswith("X"):
            return "USAGE_Intermediate"
         # there is also a USAGE_Intermediate, but the exchange
         # symmetry declarations below might not actually match that.
         raise Exception("Usage class for '{}' not defined.".format(Name))
   def MakeSymDecl(NameAndDomain):
      Name,Domain,Spins = NameAndDomain.split(":")
      if GetUsageClass(Name) == "USAGE_Intermediate":
         # these are not yet coded (technically, there is code to determine
         # the intermediate properties, but they are not propagated to this place)
         return ""
      # just make simple double exchange symmetries for a start.
      # (here we assume that all tensors have them)
      n = len(Domain)
      assert(n%2 == 0)
      if len(set(Spins[:n//2])) != 1:
         # they still can have some symmetries, but only for
         # exchanging slots of compatible spin. Leave them alone for now.
         return ""
      if n//2 >= 2 and len(set(Domain[:n//2])) == 1 and len(set(Domain[n//2:])) == 1:
         return "0123456789"[:n//2] + "/" + "0123456789"[n//2:n]
      return ""
   Decls = []
   def CombineDomainAndSpin(Domains,Spins):
      Out = []
      for (d,s) in zip(Domains, Spins):
         if s == 'A':
            d = d.upper()
         Out.append(d)
      return "".join(Out)
   for (iTensor,NameAndDomain) in TensorNames.items():
      Name,Domain,Spins = NameAndDomain.split(":")
      if MethodSpinClass == "R":
         Spins = Spins.replace("B","_")
      #SymWarning = " /* WARNING: Symmetry decl likely wrong!*/" if GetUsageClass(Name) == "USAGE_Intermediate" else ""
      Decls.append(' /*{:3}*/  {{"{}", "{}", "{}", {}}}'.format(
         iTensor, Name, CombineDomainAndSpin(Domain,Spins),
         MakeSymDecl(NameAndDomain), GetUsageClass(Name)))
   Emit("FTensorDecl TensorDecls[{}] = {{", len(Decls)); Ind()
   Emit(",\n".join(Decls))
   UnInd(); Emit("};")

   def GetTensorIndex(NameAndIndices):
      Name, Domain, Spins = ToNameAndDomain(NameAndIndices).split(":")
      return TensorIndices["{}:{}:{}".format(Name, Domain, Spins)]

#struct FEqInfo
#{
   #char const
      #*pCoDecl; // <- note that these are only used to determine the
                #//    permutations. The names of the indices do not matter.
   #double
      #Factor;
   #unsigned
      #nTerms;
   #int
      #// first one is lhs, the others are rhs.
      #iTerms[1+EQINFO_MaxRhs];
#};
   nKept = {}
   Annotate = True
   for (SetName, Cos) in EqSets:
      Eqs = []
      for Co in Cos:
         Eqs.append('{{"{}", {}, {}, {{{}}}}}~'.format(
            MakeCoStr(Co, IndexToBase), FloatToStr(Co.Factor), 1 + len(Co.Rhs),
            ",".join(map(str,map(GetTensorIndex, [Co.Lhs] + Co.Rhs)))))
         if Annotate:
            Eqs[-1] = "{:<60} // {}".format(Eqs[-1], Co)
      Emit("FEqInfo Eqs{}[{}] = {{", SetName, len(Eqs)); Ind()
      Eqs[-1] = Eqs[-1].replace("~","" if not Annotate else " ")
      Emit("\n".join(o.replace("~",",") for o in Eqs))
      UnInd(); Emit("};")
      nKept[SetName] = len(Eqs)
   BeginBlock("static void GetMethodInfo(FMethodInfo &Out)")
   Emit('Out = FMethodInfo();')
   Emit('Out.pName = "{}";', MethodName)
   Emit('Out.pSpinClass = "{}";', {"R":'restricted', "U":'unrestricted', "G":'generalized'}[MethodSpinClass])
   Emit('Out.pTensorDecls = &TensorDecls[0];')
   Emit('Out.nTensorDecls = {};', len(Decls))
   for (SetName,Cos) in EqSets:
      Emit('Out.Eqs{} = FEqSet(&Eqs{}[0], {}, "{}/{}");', SetName, SetName, nKept[SetName], MethodName, SetName)
   EndBlock()

   return CodeGen._OutputLines

def Indent(In,Ind):
   Out = []
   for Line in In.split("\n"):
      s = Line.rstrip()
      if s != "":
         Out.append(Ind + s)
      else:
         Out.append("")
   return "\n".join(Out)


def ConvertFiles(FileNameOut, NamespaceName, MethodName, FileNamesIn, SpinClass="R"):
   def ReadFile(s):
      with open(s,"r") as File:
         return utf8_decode(File.read())
   def ReadContractions(FileName):
      def RemoveComments(s):
         ii = s.find("//")
         if ii == -1:
            return s
         else:
            return s[:ii].rstrip()
      Lines = [RemoveComments(o) for o in ReadFile(FileName).splitlines()]
      Cos = [FContraction(o) for o in Lines if o.strip() != ""]
      return Cos

   EqSets = [(o[0], ReadContractions(o[1])) for o in FileNamesIn]
   LinesOut = ConvertToSrci(MethodName, EqSets, SpinClass);

   TextOut = "\n".join(LinesOut) + "\n"
   TextOut = "namespace {} {{\n{}}} // namespace {}".format(
      NamespaceName, Indent(TextOut,"   "), NamespaceName)
   TextOut += "\n// kate: syntax c++ "

   with open(FileNameOut, "w") as File:
      File.write(utf8_encode(TextOut))
      print("* wrote '{}'".format(FileNameOut))


def _MainConvertItfEncResult():
   d = "eqs/"
   FileRes = d + "EqCcsd-Res.itfaa"
   #FileRes = "/home/cgk/dev/network_factorize/_factor_result.itfaa"
   ConvertFiles('/home/cgk/dev/srci/EqCcsd-R.inl',
      "EqCcsdR", "CCSD",
      [("Res",FileRes),
       ("A2r",d+"EqCcsd-A2r.itfaa"),
       ("R2a",d+"EqCcsd-R2a.itfaa"),
       ("Nrm",d+"EqCcsd-Nrm.itfaa"),
       ("EHy2",d+"EqCcsd-EHy2.itfaa"),])
       #("Rdm1",d+"EqCcsd-Rdm1.itfaa")])

if __name__ == "__main__":
   _MainConvertItfEncResult()
