# Copyright (c) 2013-2019 Gerald Knizia
# 
# This file is part of the 'cc_eqs_pv' program
# (See https://sites.psu.edu/knizia/)
# 
# cc_eqs_pv is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
# 
# cc_eqs_pv is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with cc_eqs_pv (LICENSE_GPL_v3.txt). If not, see
# http://www.gnu.org/licenses/

from __future__ import print_function
import itertools as it
import numpy as np
from scipy import linalg as la

# number of electrons in a closed-shell orbital.
ORB_OCC = 2


def FindCycles(l,IncludeLength1=True):
   """return list of cycles of a permutation. By default includes cycles of length 1."""
   # c/p'd from http://corte.si/posts/code/cyclesort/index.html
   seen = set()
   cycles = []
   for i in range(len(l)):
      if (not i in seen) and (IncludeLength1 or i != l[i]):
         cycle = []
         n = i
         while 1:
            cycle.append(n)
            n = l[n]
            if n == i:
               break
         seen = seen.union(set(cycle))
         cycles.append(list(reversed(cycle)))
   return cycles


def InvertPerm(l):
   """return inverse permutation of a given permutation"""
   out = len(l)*[0]
   for (io,o) in enumerate(l):
      out[o] = io
   return out


def _GetPermWeight(p_):
   return (-1)**len(p_) * (-ORB_OCC)**len(FindCycles(p_))


def _MakeMMatrix(Perms):
   """construct the matrix 'M' of the cc-closed paper which represents the
   overlap of E-operators with permuted indices occupied indices applied
   to the reference function. Concretely,
   
       M[ip,iq] = <Phi| E^{p(ijk..)}_{abc..} E^{abc..}_{q(ijk..)} |Phi>
       
   The order of permutations (and thereby the indexing) comes from the Perms
   argument.
   E.g., for closed-shell case (ORB_OCC=2) and two indices, with Perms = [(0,1), (1,0)],
   this generates and returns the matrix

        [ <Phi|E^{ij}_{ab} E^{ab}_{ij}|Phi>    <Phi|E^{ji}_{ab} E^{ab}_{ij}|Phi> ]   [ 4   2 ]
   M =  [                                                                        ] = [       ]
        [ <Phi|E^{ij}_{ab} E^{ab}_{ji}|Phi>    <Phi|E^{ji}_{ab} E^{ab}_{ji}|Phi> ]   [ 2   4 ]
   """
   Np = len(Perms)
   M = np.zeros((Np,Np))
   for (ip,p) in enumerate(Perms):
      for (iq,q) in enumerate(Perms):
         invp = InvertPerm(p)
         pq = [q[invp[o]] for o in range(len(p))]
         M[ip,iq] = _GetPermWeight(pq)
   return M


def PrintCaption(Caption, Perms, Rank):
   print("\n\n *** {}: RANK = {} ***".format(Caption, Rank))
   print("\n Target Permutation List:")
   for (ip, Perm) in enumerate(Perms):
      assert(len(Perm) == Rank)
      print(" :{:4}    {:20} -> cycles: {}".format(ip, str(Perm), FindCycles(Perm)))


def MakeContraVariantProjectionTrafos(Perms, Rank):
   PrintCaption("MAKE CONTRAVARIANT PROJECTIONS", Perms, Rank)

   Np = len(Perms)  # number of permutations
   M = _MakeMMatrix(Perms)
   print(" Permutational Overlap matrix:\n{}".format(M))
   assert(M.shape == (Np,Np))

   # make target vector k1 for solving for the contravariant linear combinations
   # of permutations r0. k1 is all zeros, except for a 1 on the identity permutation.
   # What we now search is the linear combination r0 such that
   #
   #     ||M * r0 - vIdPerm|| -> min.
   #
   # I.e., such that we get the closest possible relation between rhs
   # combinations (represented by r0) and the contravariant residuals (represented by vIdPerm).
   vIdPerm = np.zeros(Np)
   IdPerm = tuple(range(Rank))
   vIdPerm[Perms.index(IdPerm)] = 1.

   # solve for the target linear combinations
   r0 = np.linalg.lstsq(M,vIdPerm, rcond=1e-10)[0]
   # set all near-zero values (due to rounding) to exactly zero.   
   r0[np.abs(r0)<1e-10] = 0.

   if 1:
      # normalize on identity component: scale all permutation factors in such a
      # way that the contravariant projection applied to its co-variant base operator
      # gives unit overlap:
      #
      #    <{\tilde E}^{ijk}_{abc}  E^{abc}_{ijk}> = 1
      fid = np.dot(np.dot(M,vIdPerm),r0)
      r0 /= fid
      print(" Rescaling contravariant permutation factors by: {:.15f}".format(1./fid))

   ContraVariantPrefactors = r0

   # Make a list of the entries (p, coeff) mapping all index permutations
   # to their respective coefficients. This is formatted as Python code and
   # included as a constant in the main file.
   print(" Output contravariant projection permutation factors:")
   L = []
   for (ip, (p, coeff)) in enumerate(zip(Perms, ContraVariantPrefactors)):
      sPerm = ", ".join(str(o) for o in p)
      Entry = "(({}): {:.15f})".format(sPerm, coeff)
      L.append(Entry)
      print(" :{:4}    {:20} -> {:20.15f}".format(ip, str(p), coeff))
   return ", ".join(L)


def MakeResidualCleanupCode(Perms, Rank):
   PrintCaption("MAKE RESIDUAL CLEANUP TRANSFORMATION", Perms, Rank)

   Np = len(Perms)  # number of permutations
   M = _MakeMMatrix(Perms)

   print(" Permutational Overlap matrix:\n{}".format(M))
   assert(M.shape == (Np,Np))

   ew,ev = la.eigh(M)
   print(" Eigenvalues:\n{}".format(ew))

   NullSpace = [ev[:,i] for i in range(len(ew)) if np.abs(ew[i])<1e-10]
   NullSpace = np.array(NullSpace).T
   print("NullSpace dimension: %s" % (NullSpace.shape,))
   
   PNull = np.dot(NullSpace, NullSpace.T)
   PNull /= 1./float(Np)
   print("NullSpace projector:\n%s" % (PNull))
   
   DefSpace = [ev[:,i] for i in range(len(ew)) if np.abs(ew[i])>=1e-10]
   DefSpace = np.array(DefSpace).T
   PDef = np.dot(DefSpace, DefSpace.T)
   PDef /= 1./float(Np)
   print("DefSpace projector:\n%s" % (PDef))
   
   def MakeSrciIncludeCode():
      """this generates the residual cleanup code used in the srci reference implementation"""
      Lines = []
      def Emit(s):
         Lines.append("      " + s)
      PermedT = []
      for p_ in Perms:
         # I wonder if the permutations come in the same order in pypy and CPython...
         p = InvertPerm(p_)
         abc = "abcdefgh"[:Rank]
         ijk = "ijklmnop"[:Rank]
         PermedT.append("&t(%s,%s)" % (",".join(abc[o] for o in p), ",".join(ijk)))
      Emit("static double const s_CleanPrj%s[%s*%s] = {%s};" % (Np,Np,Np, ",".join("%.0f" % o for o in PDef.flatten())))
      Emit("double *p[%s] = {%s};" % (len(PermedT), ", ".join(PermedT)))
      Emit("double c[%s], v[%s];" % (len(PermedT), len(PermedT)))
      Emit("for (size_t ip = 0; ip < %s; ++ ip) v[ip] = *p[ip];" % Np)
      Emit("ct::Mxv(c,1, s_CleanPrj{0},{0},1, v,1, {0},{0}, false, 1./{0}.);".format(Np))
      Emit("for (size_t ip = 0; ip < %s; ++ ip) *p[ip] = c[ip];" % Np)
      print("\n --- SrCi Residual Cleanup Code:\n\n{}\n ---\n".format("\n".join(Lines)))

   def MakeLatexIncludeCode():
      """this generates the latex code used in the cc-closed article manuscript"""
      Lines = []
      def Emit(s):
         Lines.append("      " + s)
      def MakePermedSymbols(TensorName):
         PermedT = []
         for p_ in Perms:
            # I wonder if the permutations come in the same order in pypy and CPython...
            p = InvertPerm(p_)
            abc = "abcdefgh"[:Rank].upper()
            ijk = "ijklmnop"[:Rank].upper()
            PermedT.append("%s^{%s}_{%s}" % (TensorName, "".join(abc[o] for o in p), "".join(ijk)))
         return PermedT
      RowSepV = r" \\[.8ex] "
      RowSepT = r" \\[\vtab] "
      
      def MakeVec(Entries):
         return r"\left[\begin{{array}}{{c}} {} \end{{array}}\right]".format(RowSepV.join(str(o) for o in Entries))
      def MakeMatrix(M, Lead0 = "         "):
         (nRows, nCols) = M.shape
         L = []
         L.append(r"\left[\begin{array}{%s}" % (nCols * "r"))
         for iRow in range(nRows):
            
            if iRow == 0:
               Lead = Lead0 + r" "*len(RowSepT)
            else:
               Lead = Lead0 + RowSepT
            ColData = ["{:>8}".format(str(o)) for o in M[iRow,:]]
            L.append(Lead + " & ".join(ColData))
         L.append(Lead0 + r"\end{array}\right]")
         return "\n".join(L)
         
      Eq = r"""
      \newlength{\vtab}\setlength{\vtab}{1.0ex}
      \begin{align*}
         %s
         := \frac{1}{%s}
         %s
         \cdot
         %s
      \end{align*}
      """ % (MakeVec(MakePermedSymbols("r")), str(Np), MakeMatrix(np.rint(PDef).astype(int)), MakeVec(MakePermedSymbols("r")))
      print("\n\n{}\n\n".format(Eq))
      
      pass

   
   MakeSrciIncludeCode()
   #MakeLatexIncludeCode()

   
   

def main():
   ContraVariantLeftPermPrefactors = []
   for Rank in [1,2,3,4]:
   #for Rank in [1,2,3]:
      Perms = list(it.permutations(range(Rank)))
      L = MakeContraVariantProjectionTrafos(Perms, Rank)
      ContraVariantLeftPermPrefactors.append("   {}: {}".format(Rank, L))
      
      MakeResidualCleanupCode(Perms, Rank)
   print("\n --- cc-eqs-pv Covariant Projection Data Include:\n")
   print("_DefaultCovariantLeftPermAndFactors = {{\n{}\n}}".format("\n".join(ContraVariantLeftPermPrefactors)))
   


if __name__ == "__main__":
   np.set_printoptions(linewidth=881123, precision=5)
   main()
